!(function (factory) {
	if (typeof define === 'function' && define.amd) {
		// AMD (Register as an anonymous module)
		define(['lodash'], factory);
	} else if (typeof exports === 'object') {
		// Node/CommonJS
		module.exports = factory(require('lodash'));
	} else {
		/* eslint-disable no-undef */
		factory(_);
		/* eslint-enable no-undef */
	}
}(function (_) {
	// this is for backward compatibility, weird but required
	var legacyBoolean = false;
	var translate = {
		"_i_": function (pr) {
			pr = parseInt(pr);

			if (!isNaN(pr))
				return pr;
		},
		"_s_": function (pr) {
			return pr.toString();
		},
		"_id": function (pr) {
			return pr.toString();
		},
		"_f_": function (pr) {
			pr = parseFloat(pr);

			if (!isNaN(pr))
				return pr;
		},
		"_t_": function (pr) {
		},
		"_dt": function (pr) {
			var t = Date.parse(pr);
			if (!isNaN(t))
				return new Date(t);
			else if (!isNaN(parseInt(pr)))
				return new Date(parseInt(pr));
			else if (_.isDate(pr))
				return pr;
		},
		"_b_": function (pr) {
			if (pr === true || pr === 1 || pr === "true" || pr === "1")
				return legacyBoolean?1:true;
			if (pr === false || pr === 0 || pr === "false" || pr === "0" || pr === null || pr === "null")
				return legacyBoolean?0:false;
		}
	};

	function sortfix(obj) {
		var nobj = {};
		_.forOwn(obj, function (v, k) {
			nobj[k] = parseInt(v);
		});
		return nobj;
	}

	function queryfix(obj, opts) {
		if (!obj || !_.isPlainObject(obj))
			return null;

		var nobj = {};
		_.forOwn(obj, function (v, k) {
			// query can use dot notation for names
			// last component should refer to actual type
			var prefix = k.indexOf("_") !== -1 ? _.last(k.split('.')).match(/^(_..)/) : null;
			if (prefix)
				prefix = prefix[1];

			if (prefix === "_t_")
				return;

			var exists = _.has(v, "$exists");

			if (!exists && prefix && translate[prefix]) {
				// object meand op, like {$gt:5,$lt:8}
				if (_.isPlainObject(v)) {
					var no = { };
					_.forOwn(v, function (val, op) {
						// op value is array {$in:[1,2,4]}
						if (Array.isArray(val)) {
							no[op] = _.reduce(val, function (memo, a) {
								try {
									memo.push(translate[prefix](a));
								} catch (e) { /**/ }

								return memo;
							}, []);
						} else if (op === '$regex' && _.isRegExp(val)) { // special case
							no[op] = val;
						} else {
							try {
								no[op] = translate[prefix](val);

								if (no[op] === undefined) {
									no[op] = val;
								}
							} catch (e) {
								no[op] = val;
							}
						}
					});
					nobj[k] = no;
				} else if (prefix === "_s_" && _.isRegExp(v)) { // special case
					nobj[k] = v;
				} else {
					// plain value then
					try {
						nobj[k] = translate[prefix](v);

						if (nobj[k] === undefined) {
							nobj[k] = v;
						}
					} catch (e) {
						nobj[k] = v;
					}
				}
			} else {
				if (_.isPlainObject(v)) {
					nobj[k] = queryfix(v, opts);
				} else if (Array.isArray(v) && (k === "$or" || k === "$and")) {
					var res = _.reduce(v, function (memo, a) {
						var r = queryfix(a, opts);

						if (_.size(r)) {
							memo.push(r);
						}

						return memo;
					}, []);

					if (res.length) {
						nobj[k] = res;
					}
				} else {
					nobj[k] = v;
				}
			}
		});
		return nobj;
	}

	function datafix(obj,opts) {
		var nobj = obj;

		_.forOwn(obj, function (v, k) {
			if (_.isFunction(v) || _.isUndefined(v))
				return;

			var prefix = null;
			if (k.length > 2 && k[0] == "_")
				prefix = k.substr(0,3);

			if (prefix && translate[prefix]) {
				var nv;
				try { nv = translate[prefix](v); } catch (e) { /**/ }
				if (_.isUndefined(nv)) {
					if (opts && opts.strict)
						throw new Error("Wrong field format: "+k);
					delete nobj[k];
				} else if (nv!==v)
					nobj[k] = nv;
			} else if (_.isPlainObject(v) || Array.isArray(v)) {
				datafix(v,opts);
			}
		});
		return nobj;
	}

	return {
		queryfix:queryfix,
		datafix:datafix,
		data:datafix,
		query:queryfix,
		sort:sortfix,
		register:function (prefix, transform) {
			translate[prefix]=transform;
		},
		configure:function (cfg) {
			legacyBoolean = !!cfg.legacyBoolean;
		}
	};
}));
